/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "blade-config.h"
#include <iostream>
#include <cstdlib>
#include <algorithm> 
#include <QMouseEvent>
#include <QToolTip>
#include "MetricGraphicsView.h"
#include "TraceEventRect.h"
#include "TraceEventGroupWidget.h"
#include "Service.h"
#include "TraceEventGroupWidget.h"


extern TraceEventGroupWidget * tegw;

MetricGraphicsView::MetricGraphicsView ( QGraphicsScene* scene, MetricLine * _metric_line) : QGraphicsView( scene ), metric_line(_metric_line)
{
    zooming = false;
    pressed = false;
    x_treshold = 15;
    setMouseTracking(true);
    start_position = QPoint(0, 0);
    end_position = QPoint(0, 0);
    
    setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
    setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
        
    zoom_rect = calculateRect( start_position, end_position );
    zoom_window = calculateZoomF();
    
    zoom =  new  QGraphicsRectItem();
    cursor = new QGraphicsLineItem();
    
}; 
    
   
void
MetricGraphicsView::updateScene()
{
    zoom =  new  QGraphicsRectItem();
    cursor = new QGraphicsLineItem();
    
    QPen pen;
    pen.setColor( QColor( 0,0,0,100));
    QBrush brush;
    brush.setColor( QColor( 0,0,0,100));
    brush.setStyle(Qt::SolidPattern);
    
    
    QPen cursor_pen;
    cursor_pen.setColor( QColor( 0,0,0,255));
    cursor_pen.setWidth(2);
    cursor->setPen( cursor_pen );
    
    
    zoom->setPen( pen );
    zoom->setBrush( brush );
    
    zoom->setRect( zoom_rect );
    
    scene()->addItem( zoom );
    scene()->addItem( cursor );
}
    
QPointF&
MetricGraphicsView::getZoomF()
{
    return zoom_window;
}
    
QPointF
MetricGraphicsView::getCursorPosition()
{
    return QPointF(
        (double)end_position.x()/ (double)width()
        ,0
    );
}
  
void
MetricGraphicsView::setCursorPosition(QPointF& _p)
{
    cursorPosition = QPoint( _p.x() * width(),0) ;
    cursor->setLine(cursorPosition.x(),0, cursorPosition.x(), height() );
}  
  
  
QRect 
MetricGraphicsView::calculateRect( const QPoint& _s, const QPoint& _f)
{
    QSize _size = QSize(abs( _s.x() - _f.x()), height());
    
    QPoint _p = QPoint(std::min( _s.x(), _f.x()), 0 ) ;
    QRect _r = QRect( _p, _size);
    return _r;
}
    
    
    
QPointF
MetricGraphicsView::calculateZoomF( )
{
    QPointF _p = QPointF(
        (double)zoom_rect.x()/ (double)width()
        ,
        (double)(zoom_rect.x() + zoom_rect.width())/ (double)width()
    );

    return _p;
}
     
    
void 
MetricGraphicsView::mousePressEvent(QMouseEvent * _event)
{
    if (_event->button() == Qt::RightButton)
        return;
    int _x = _event->x();
    int _y = _event->y();
    if ( _y < 0 || _y > height())
        return;
    if ( _x < 0 || _x > width())
        return;
    
    zooming = false;
    pressed = true;
    start_position.setX( _x );
    start_position.setY( _y );
    end_position.setX( _x );
    end_position.setY( _y );
    zoom_rect = calculateRect( start_position, end_position );
    zoom->setRect(zoom_rect);
}

void
MetricGraphicsView::mouseMoveEvent(QMouseEvent * _event)
{

    int _x = _event->x();
    int _y = _event->y();
    if ( _y < 0 || _y > height())
        return;
    
    if ( _x < 0 || _x > width())
    {
        mouseReleaseEvent( _event);
        return;
    }
    
    if (!zooming && pressed)
        if ( std::abs(_x - start_position.x()) > x_treshold)
            zooming = true;
        else
            return;
    
    
    end_position.setX( _x );
    end_position.setY( _y );     
    

    if (!zooming)
    {
        uint64_t _s = z_filter->get_time_in_window( (double)_x/(double)(width()));
        QToolTip::showText(mapToGlobal(QPoint(_x, height()/2)), QString("Value at %1: %2").arg(_s).arg(metric_line->evaluate(_s)) );     
        emit cursorMoved(getCursorPosition());
        return;
    }
    else
    {       
        zoom_rect = calculateRect( start_position, end_position );
        zoom_window = calculateZoomF();
        zoom->setRect(zoom_rect);
        emit zoomingSelection();        
    }
}

void 
MetricGraphicsView::mouseReleaseEvent(QMouseEvent * _event)
{
    int _x = _event->x();
    int _y = _event->y();
    if (!zooming)
    {

        start_position.setX( 0 );
        start_position.setY( 0 );
        end_position.setX( 0 );
        end_position.setY( 0 );
        pressed = false;
        return;  
    }
    pressed = false;
    if ( _x < 0 )
    {
        _x =0 ;
    }  
    if (  _x > width())
    {
        _x = width() ;
    }  
    end_position.setX( _x );
    end_position.setY( _y );     
    zoom_rect = calculateRect( start_position, end_position );   
    zoom_window = calculateZoomF();
    
    zooming = false;
    start_position.setX( 0 );
    start_position.setY( 0 );
    end_position.setX( 0 );
    end_position.setY( 0 );
    zoom_rect = calculateRect( start_position, end_position );
    zoom->setRect(zoom_rect);
    
    emit zoomChanged();
    
}


     
void
MetricGraphicsView::showToolTip( QString _tip)
{
    QToolTip::showText(mapToGlobal(end_position),_tip );  
}
