/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "blade-config.h"
#include <iostream>

#include <QMouseEvent>
#include <QToolTip>
#include "OverviewGraphicsView.h"
#include "TraceEventRect.h"
#include "BladeUtils.h"
#include "Service.h"

#include "PluginServices.h"
extern cubepluginapi::PluginServices* blade_service;

OverviewGraphicsView::OverviewGraphicsView ( QGraphicsScene* scene,  ZoomFilter * _z_filter, Trace * _trace) : QGraphicsView( scene ), z_filter( _z_filter), trace(_trace)
{
    zooming = false;
    setMouseTracking(true);
    start_position = QPoint(0, 0);
    end_position = QPoint(0, 0);
    
    zoom_rect = calculateRect( start_position, end_position );
    zoom_window = calculateZoomF();
    
    
}; 
    
   
void
OverviewGraphicsView::updateScene()
{
    zoom =  new  QGraphicsRectItem();
    
    QPen pen;
    pen.setColor( QColor( 0,0,0,255));
    pen.setWidth(2);
    QBrush brush;
    brush.setColor( QColor( 0,0,0,127));
    brush.setStyle(Qt::SolidPattern);
    
    
    zoom->setPen( pen );
    zoom->setBrush( brush );
    
    zoom->setRect( zoom_rect );
    
    scene()->addItem( zoom );
}
    
QPointF&
OverviewGraphicsView::getZoomF()
{
    return zoom_window;
}

  
QRect 
OverviewGraphicsView::calculateRect( const QPoint& _s, const QPoint& _f)
{
    QSize _size = QSize(abs( _s.x() - _f.x()), height());
    
    QPoint _p = QPoint(std::min( _s.x(), _f.x()), 0 ) ;
    QRect _r = QRect( _p, _size);
    return _r;
}
    
    
    
QPointF
OverviewGraphicsView::calculateZoomF( )
{
    QPointF _p = QPointF(
        (double)zoom_rect.x()/ (double)width()
        ,
        (double)(zoom_rect.x() + zoom_rect.width())/ (double)width()
    );
  
    return _p;
}
     
    
void 
OverviewGraphicsView::mousePressEvent(QMouseEvent * _event)
{
    int _x = _event->x();
    int _y = _event->y();
    if ( _y < 0 || _y > height())
        return;
    if ( _x < 0 || _x > width())
        return;
    
    zooming = true;
    start_position.setX( _x );
    start_position.setY( _y );
    end_position.setX( _x );
    end_position.setY( _y );
    zoom_rect = calculateRect( start_position, end_position );
    zoom->setRect(zoom_rect);
}

void 
OverviewGraphicsView::mouseMoveEvent(QMouseEvent * _event)
{

    int _x = _event->x();
    int _y = _event->y();
    if ( _y < 0 || _y > height())
        return;
    
    if ( _x < 0 || _x > width())
    {
        mouseReleaseEvent( _event);
        return;
    }
    
    end_position.setX( _x );
    end_position.setY( _y );     
/*    
    uint64_t _s = z_filter->get_time_in_window( (double)_x/(double)width());
    std::pair<double, uint16_t> _s_in_seconds = trace->convert_to_seconds( z_filter->get_zoom_start() );
    std::pair<double, uint16_t> _f_s_in_seconds = trace->convert_to_seconds( z_filter->get_zoom_end() );
    std::pair<double, uint16_t> _p_in_seconds = trace->convert_to_seconds( _s );
    blade_service->updateValueWidget(   NULL,
                                        _s_in_seconds.first  ,
                                        _f_s_in_seconds.first ,
                                        _p_in_seconds.first,
                                        0.,
                                        0.,
                                        0.,
                                        0.,
                                        0.
                                    );
    */
    
    if (!zooming)
    {
        TraceEventRect * _rect = dynamic_cast<TraceEventRect*>( itemAt(_x, _y));
        if (_rect != NULL )
        {
            QString _tip = 
            "Region name: " +
            BladeUtils::shortCutRegionName(_rect->getRegionName())+"\n"
            "Start time : "+
            _rect->getStartTime() + "\n"
            "Duration : " +
            _rect->getDuration() + "\n"
            "Call Path: \n" +
            _rect->getCallPath();
                
            QToolTip::showText(mapToGlobal(QPoint(_x, _y)),_tip );     
        }else
        {
            QToolTip::hideText();
        }
        return;
    }
    else
    {       
        
        zoom_rect = calculateRect( start_position, end_position );
        zoom_window = calculateZoomF();
        zoom->setRect(zoom_rect);
        emit zoomingSelection();        
    }
    


}

void 
OverviewGraphicsView::mouseReleaseEvent(QMouseEvent * _event)
{
    if (!zooming)
        return;  
    int _x = _event->x();
    int _y = _event->y();
    if ( _x < 0 )
    {
        _x =0 ;
    }  
    if (  _x > width())
    {
        _x = width() ;
    }  
    end_position.setX( _x );
    end_position.setY( _y );     
    zoom_rect = calculateRect( start_position, end_position );   
    zoom_window = calculateZoomF();
    
    zooming = false;
    emit zoomChanged();
    
}


void
OverviewGraphicsView::changeZoom(QPointF _pp)
{
    zoom_rect = calculateRect( QPoint(_pp.x()*width(), 0), QPoint(_pp.y()*width(),0) );  
    zoom->setRect(zoom_rect);
}

void
OverviewGraphicsView::showToolTip( QString _tip)
{
    QToolTip::showText(mapToGlobal(end_position),_tip );  
}
