/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "blade-config.h"
#include <iostream>
#include <utility>

#include "Trace.h"
#include "Service.h"
#include "TraceGlobals.h"


TraceGlobals * trace_globals;



Trace::Trace()
{
    
    mpi_collective_groups = NULL;
    mpi_p2p_groups = NULL;
    omp_parallel_groups = NULL;
    trace_globals = NULL;
    
    callpath_filter_active = false;
    selection_filter_active =false;
    trace_globals = new TraceGlobals();

}


Trace::~Trace()
{
    for (std::map<uint64_t, TraceEvent*>::iterator iter=entries.begin(); iter!= entries.end(); ++iter)        
        delete (iter->second);
    delete mpi_collective_groups;
    delete mpi_p2p_groups;
    delete omp_parallel_groups;
    delete trace_globals;
}

void
Trace::set_events( std::map<uint64_t, TraceEvent *>& _trace)
{
    entries = _trace;
    number_of_locations = entries.size();
}
    
void
Trace::set_max_levels( std::map<uint64_t, int32_t> _max_levels )
{
    max_levels= _max_levels;
}   
    
    
void
Trace::set_system_tree(std::vector<uint64_t>& _root_stn, 
                       std::map<uint64_t, SystemTreeNodeClass>& _stn, 
                       std::map<uint64_t, LocationGroupClass>& _lg, 
                       std::map<uint64_t, LocationClass> & _locs )
{
    locations_classes = _locs;  
    lg_classes = _lg;  
    stn_classes = _stn;  
    root_stn_classes = _root_stn;  
}
   
   
void
Trace::set_metrics( std::vector< MetricMemberClass >& _metric_members,
             std::vector< MetricClassClass >& _metric_classes,
             std::vector< MetricInstanceClass >& _metric_instances )
{
    metric_members = _metric_members;
    metric_classes = _metric_classes;
    metric_instances = _metric_instances;
}
   
    
void
Trace::init( uint64_t _s, uint64_t _f, uint64_t _tr, uint64_t _go, uint64_t _tl,  std::vector<RegionClass>& _regions,  std::vector<CallPathClass> _calltree )
{
    start_timestamp = _s;
    end_timestamp = _f ;
    timer_resolution = _tr;
    global_offset = _go;
    trace_length = _tl;
    regions = _regions;
    calltree = _calltree;
 }


   
void
Trace::print()
{
        for (std::map<uint64_t, TraceEvent*>::iterator iter=entries.begin(); iter!= entries.end(); ++iter)        
        {
          print(iter->first);
        }
}
 
void
Trace::print(uint64_t _location)
{
    std::map<uint64_t, TraceEvent*>::iterator iter=entries.find(_location);
    if (iter ==entries.end())
        return ;
    (iter->second)->print( 0 );
    
}

std::pair<double, uint16_t> 
Trace::convert_to_seconds( uint64_t _timestamp ) const
{
    double _in_sec = (double)_timestamp / (double)timer_resolution;
    std::pair<double, uint16_t> _human_readable = std::make_pair( _in_sec, 0);
    while (_human_readable.first > 0. && _human_readable.first < 0.1)
    {
        _human_readable.first *= 1000;
        _human_readable.second += 3;
    }
    return _human_readable;
}
    
     

void
Trace::clear_trace_view(std::vector<TraceEvent* >&_trace_view)
{    
    _trace_view.clear();
}

std::vector<TraceEvent* >&
Trace::get_trace_view(uint64_t _location, ZoomFilter * _z_f, TraceEventFilter * _cf_f, SelectionFilter * _s_f, ResolutionFilter * _r_f, LocationFilter * _l_f, RoleFilter * _role_f, RegionPropertiesFilter * _region_p_f)
{
    uint64_t _m_location = bladeservices::master_id( _location );
    std::map<uint64_t, TraceEvent*>::iterator iter=entries.find(_m_location );
    if (iter == entries.end())
    {
        return trace_view[_location];
    }
    
    clear_trace_view(trace_view[_location]);  
    
    
    TraceEvent * _entry = iter->second;
    for (size_t i=0; i< _entry->num_sub_calls(); ++i)
    {
         TraceEvent * sub_entry = _entry->get_sub_call(i);
         add_to_view(trace_view[_location], sub_entry, _z_f, _cf_f, _s_f, _r_f, _l_f, _role_f, _region_p_f);
    }
    return trace_view[_location];
}



void
Trace::add_to_view(std::vector<TraceEvent* >&_trace_view, TraceEvent *_entry, ZoomFilter * _z_f, TraceEventFilter * _cf_f,  SelectionFilter * _s_f, ResolutionFilter * _r_f, LocationFilter * _l_f, RoleFilter * _role_f, RegionPropertiesFilter * _region_p_f)
{
    _entry->set_visible( false );
    _entry->set_exclusive( false );
    if (_z_f->apply(_entry) == FILTER_FAIL)
    {
        return ;
    }
    FilterResult _l_f_result = _l_f->apply(_entry);
    if ( _l_f_result == FILTER_FAIL )
    {
        return ;
    }  
    FilterResult _region_p_f_result = _region_p_f->apply(_entry);
    if ( _region_p_f_result == FILTER_FAIL)
    {
        return ;
    } 
    FilterResult _role_f_result = _role_f->apply(_entry);
    if ( _role_f_result == FILTER_FAIL)
    {
        return ;
    } 
    FilterResult _c_f_result = _cf_f->apply(_entry); 
    if (_c_f_result == FILTER_FAIL )
    {
        return ;
    }  
    FilterResult _r_f_result = _r_f->apply(_entry);
    if (  _r_f_result == FILTER_FAIL && (_entry->get_parent()->is_visible() && _entry->get_parent()->get_location() == _entry->get_location()))
    {
        return ;
    }  
    FilterResult _s_f_result = FILTER_FAIL;
   if ( ((_role_f_result == FILTER_OK) && (_region_p_f_result == FILTER_OK) &&( _c_f_result == FILTER_OK ||  _c_f_result== FILTER_DIRECT_CHILDREN ))   && _l_f_result ==  FILTER_OK)
    {   
        FilterResult _s_f_result = _s_f->apply(_entry);   
        if (_s_f_result == FILTER_OK)
        {
            _entry->set_visible( true );
            _entry->set_exclusive( _c_f_result == FILTER_DIRECT_CHILDREN );
            _trace_view.push_back( _entry ); 
        }
    }
    if ( _c_f_result== FILTER_DIRECT_CHILDREN )
        return;
    if ( _region_p_f_result == FILTER_CHILDREN ||_role_f_result == FILTER_CHILDREN || _c_f_result == FILTER_CHILDREN ||_r_f_result == FILTER_OK || _s_f_result == FILTER_CHILDREN )
    {
        for (size_t i=0; i< _entry->num_sub_calls(); ++i)
        {
            TraceEvent * sub_entry = _entry->get_sub_call(i);
            add_to_view(_trace_view, sub_entry, _z_f, _cf_f,  _s_f, _r_f, _l_f, _role_f, _region_p_f);
        }
    }
}
