/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2020s                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "blade-config.h"
#include <otf2/otf2.h>

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <limits>
#include <iostream>
#include <sstream>

#include "TraceLoader.h"
#include "Service.h"
#include "TraceLoadEnterFilter.h"
#include "TraceLoadLeaveFilter.h"


TraceLoader * global_trace_loader;




static
uint32_t
find_callpath( TraceEvent * _event, uint32_t parent_callpath, uint64_t region)
{
     uint64_t _r = OTF2_UNDEFINED_CALLPATH;
     if (parent_callpath != OTF2_UNDEFINED_CALLPATH)
     {
        CallPathClass _parent_cp = global_trace_loader->calltree[ parent_callpath ];
        for (std::vector<OTF2_CallpathRef>::iterator iter = _parent_cp.subcalltree.begin(); iter != _parent_cp.subcalltree.end(); ++iter)
        {
                CallPathClass __cp = global_trace_loader->calltree[*iter];
                if ( __cp.region == region)
                {
                    _r = __cp.self;
                    break;
                }
        }
     }
     else // looking for roots
     {
             for (std::vector<CallPathClass>::iterator iter = global_trace_loader->calltree.begin(); iter != global_trace_loader->calltree.end(); ++iter)
        {
                CallPathClass __cp = *iter;
                if ( __cp.parent == OTF2_UNDEFINED_CALLPATH && __cp.region == region)
                {
                    _r = __cp.self;
                    break;
                }
        }
    }
     return _r;
}

 static OTF2_CallbackCode TraceLoader_StringDefinition(void *userData, OTF2_StringRef self, const char *string)
{
      std::string _str = string;
    if (global_trace_loader->string_defs.size()<= self)
      global_trace_loader->string_defs.resize(self+1);
     global_trace_loader->string_defs[self] = _str;
        return OTF2_CALLBACK_SUCCESS;
}



static OTF2_CallbackCode
TraceLoader_RegionDefinition( void *userData, OTF2_RegionRef self, OTF2_StringRef name, OTF2_StringRef canonicalName, OTF2_StringRef description, OTF2_RegionRole regionRole, OTF2_Paradigm paradigm, OTF2_RegionFlag regionFlags, OTF2_StringRef sourceFile, uint32_t beginLineNumber, uint32_t endLineNumber )
{

    RegionClass _rc;
    _rc.self = self;
    _rc.name = (name< global_trace_loader->string_defs.size())?global_trace_loader->string_defs[name]:"UNDEFINED";
    _rc.regionRole = regionRole;
    _rc.paradigm = paradigm;
    if (global_trace_loader->regions.size()<= self)
      global_trace_loader->regions.resize(self+1);
    global_trace_loader->regions[self] = _rc;
    return OTF2_CALLBACK_SUCCESS;
}



static OTF2_CallbackCode
TraceLoader_CallPathDefinition (void *userData, OTF2_CallpathRef self, OTF2_CallpathRef parent, OTF2_RegionRef region)
{
    global_trace_loader->callpath_filter_active = true;
    global_trace_loader->selection_filter_active = true;
    CallPathClass _cpc;
    _cpc.self = self;
    _cpc.parent = parent;
    _cpc.region = region;
    if (global_trace_loader->calltree.size()<= self)
      global_trace_loader->calltree.resize(self+1);
    global_trace_loader->calltree[self] = _cpc;
    if (parent != OTF2_UNDEFINED_CALLPATH)
        global_trace_loader->calltree[parent].subcalltree.push_back(self);
    return OTF2_CALLBACK_SUCCESS;
}


static OTF2_CallbackCode
TraceLoader_ClockProperties(  void *userData,
                              uint64_t timerResolution,
                              uint64_t globalOffset,
                              uint64_t traceLength,
                              uint64_t realtimeTimestamp  )
{
    global_trace_loader->timer_resolution = timerResolution;
    global_trace_loader->global_offset = globalOffset;
    global_trace_loader->trace_length = traceLength;
    global_trace_loader->start_timestamp = 0;
    global_trace_loader->end_timestamp = traceLength;

    return OTF2_CALLBACK_SUCCESS;
}


static OTF2_CallbackCode
TraceLoader_OmpFork (OTF2_LocationRef location,
                    OTF2_TimeStamp time,
                     void *userData,
                     OTF2_AttributeList *attributeList,
                     uint32_t numberOfRequestedThreads)
{
   std::map<uint64_t, int32_t> _subthreads = global_trace_loader->forks[ location ];
   for ( std::map<uint64_t, int32_t>::iterator iter = _subthreads.begin(); iter != _subthreads.end(); ++iter)
   {
        uint64_t _loc = (*iter).first;
        if ( global_trace_loader->calls.find(_loc) ==  global_trace_loader->calls.end())
        {
            global_trace_loader->calls[_loc].push( new TraceEvent( _loc )); // add an _root element
        }
        TraceEvent * _event = global_trace_loader->calls[location].top();
        global_trace_loader->calls[_loc].push(_event );
   }

   global_trace_loader->omp_parallel_groups->set_active( bladeservices::master_id(location), true );


    return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_OmpJoin (OTF2_LocationRef location,
                    OTF2_TimeStamp time,
                     void *userData,
                     OTF2_AttributeList *attributeList)
{
   std::map<uint64_t, int32_t> _subthreads = global_trace_loader->forks[ location ];
   for ( std::map<uint64_t, int32_t>::iterator iter = _subthreads.begin(); iter != _subthreads.end(); ++iter)
   {
        uint64_t _loc = (*iter).first;
        TraceEvent * _event = global_trace_loader->calls[_loc].top();
        global_trace_loader->calls[_loc].pop( );
   }

    global_trace_loader->omp_parallel_groups->set_active( bladeservices::master_id(location),  false );

    return OTF2_CALLBACK_SUCCESS;
}



static OTF2_CallbackCode
TraceLoader_ThreadFork (OTF2_LocationRef location,
                        OTF2_TimeStamp time,
                        void *userData,
                        OTF2_AttributeList *attributeList,
                        OTF2_Paradigm model,
                        uint32_t numberOfRequestedThreads)
{
    uint64_t _master_id = bladeservices::master_id(location);
   std::map<uint64_t, int32_t> _subthreads = global_trace_loader->forks[ _master_id ];
   for ( std::map<uint64_t, int32_t>::iterator iter = _subthreads.begin(); iter != _subthreads.end(); ++iter)
   {
        uint64_t _loc = (*iter).first;
        if ( global_trace_loader->calls.find(_loc) ==  global_trace_loader->calls.end())
        {
            global_trace_loader->calls[_loc].push( new TraceEvent( _loc )); // add an _root element
        }
        TraceEvent * _event = global_trace_loader->calls[location].top();
        global_trace_loader->calls[_loc].push(_event );
   }



    return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_ThreadJoin (OTF2_LocationRef location,
                        OTF2_TimeStamp time,
                        void *userData,
                        OTF2_AttributeList *attributeList,
                        OTF2_Paradigm model)
{
    uint64_t _master_id = bladeservices::master_id(location);
   std::map<uint64_t, int32_t> _subthreads = global_trace_loader->forks[ _master_id ];
   for ( std::map<uint64_t, int32_t>::iterator iter = _subthreads.begin(); iter != _subthreads.end(); ++iter)
   {
        uint64_t _loc = (*iter).first;
        TraceEvent * _event = global_trace_loader->calls[_loc].top();
        global_trace_loader->calls[_loc].pop( );
   }

    return OTF2_CALLBACK_SUCCESS;
}





static OTF2_CallbackCode
TraceLoader_ThreadCreate (OTF2_LocationRef location,
                          OTF2_TimeStamp time,
                          void *userData,
                          OTF2_AttributeList *attributeList,
                          OTF2_CommRef threadContingent,
                          uint64_t sequenceCount)
{
    uint64_t _master_id = bladeservices::master_id(location);
   std::map<uint64_t, int32_t> _subthreads = global_trace_loader->forks[ _master_id ];
   for ( std::map<uint64_t, int32_t>::iterator iter = _subthreads.begin(); iter != _subthreads.end(); ++iter)
   {
        uint64_t _loc = (*iter).first;
        if ( global_trace_loader->calls.find(_loc) ==  global_trace_loader->calls.end())
        {
            global_trace_loader->calls[_loc].push( new TraceEvent( _loc )); // add an _root element
        }
        TraceEvent * _event = global_trace_loader->calls[location].top();
        global_trace_loader->calls[_loc].push(_event );
   }



    return OTF2_CALLBACK_SUCCESS;
}



static OTF2_CallbackCode
TraceLoader_ThreadBegin (OTF2_LocationRef location,
                         OTF2_TimeStamp time,
                         void *userData,
                         OTF2_AttributeList *attributeList,
                         OTF2_CommRef threadContingent,
                         uint64_t sequenceCount)
{

  uint64_t _master_id = bladeservices::master_id(location);
   std::map<uint64_t, int32_t> _subthreads = global_trace_loader->forks[ _master_id ];
   for ( std::map<uint64_t, int32_t>::iterator iter = _subthreads.begin(); iter != _subthreads.end(); ++iter)
   {
        uint64_t _loc = (*iter).first;
        if ( global_trace_loader->calls.find(_loc) ==  global_trace_loader->calls.end())
        {
            global_trace_loader->calls[_loc].push( new TraceEvent( _loc )); // add an _root element
        }
        TraceEvent * _event = global_trace_loader->calls[location].top();
        global_trace_loader->calls[_loc].push(_event );
   }



    return OTF2_CALLBACK_SUCCESS;
}





static OTF2_CallbackCode
TraceLoader_ThreadEnd (OTF2_LocationRef location,
                       OTF2_TimeStamp time,
                       void *userData,
                       OTF2_AttributeList *attributeList,
                       OTF2_CommRef threadContingent,
                       uint64_t sequenceCount)
{
   uint64_t _master_id = bladeservices::master_id(location);
   std::map<uint64_t, int32_t> _subthreads = global_trace_loader->forks[ _master_id ];
   for ( std::map<uint64_t, int32_t>::iterator iter = _subthreads.begin(); iter != _subthreads.end(); ++iter)
   {
        uint64_t _loc = (*iter).first;
        TraceEvent * _event = global_trace_loader->calls[_loc].top();
        global_trace_loader->calls[_loc].pop( );
   }

    return OTF2_CALLBACK_SUCCESS;
}


















static OTF2_CallbackCode
TraceLoader_MetricMemberDefinition (void *userData, OTF2_MetricMemberRef self, OTF2_StringRef name, OTF2_StringRef description, OTF2_MetricType metricType, OTF2_MetricMode metricMode, OTF2_Type valueType, OTF2_Base base, int64_t exponent, OTF2_StringRef unit)
{
    MetricMemberClass mm;
    mm.self = self;
    mm.name = (name< global_trace_loader->string_defs.size())?global_trace_loader->string_defs[name]:"UNDEFINED";
    mm.description =  (name< global_trace_loader->string_defs.size())?global_trace_loader->string_defs[name]:"UNDEFINED";
    mm.metricType = metricType;
    mm.metricMode = metricMode;
    mm.valueType = valueType;
    mm.base = base;
    mm.exponent = exponent;
    mm.unit =  (name< global_trace_loader->string_defs.size())?global_trace_loader->string_defs[name]:"UNDEFINED";
     global_trace_loader->metric_members.push_back(mm);
    return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_MetricClassDefinition (void *userData, OTF2_MetricRef self, uint8_t numberOfMetrics, const OTF2_MetricMemberRef *metricMembers, OTF2_MetricOccurrence metricOccurrence, OTF2_RecorderKind recorderKind)
{
    MetricClassClass mc;
    mc.self = self;
    mc.numberOfMetrics = numberOfMetrics;
    for (size_t i=0; i<numberOfMetrics; ++i)
    {
        mc.metricMembers.push_back( metricMembers[i]);
    }
    mc.metricOccurrence = metricOccurrence;
    mc.recorderKind = recorderKind;

    if (global_trace_loader->metric_classes.size() <= self)
        global_trace_loader->metric_classes.resize( self+1);
    global_trace_loader->metric_classes[self] = mc;

    return OTF2_CALLBACK_SUCCESS;
}


static OTF2_CallbackCode
TraceLoader_MetricInstanceDefinition (void *userData, OTF2_MetricRef self, OTF2_MetricRef metricClass, OTF2_LocationRef recorder, OTF2_MetricScope metricScope, uint64_t scope)
{
    MetricInstanceClass mi;
    mi.self = self;
    mi.metricClass = metricClass;
    mi.recorder = recorder;
    mi.metricScope = metricScope;
    mi.scope = scope;
    if (global_trace_loader->metric_instances.size() <= self)
        global_trace_loader->metric_instances.resize( self+1);
    global_trace_loader->metric_instances[self]= mi;


    return OTF2_CALLBACK_SUCCESS;
}





static OTF2_CallbackCode
TraceLoader_ProgramBegin(   OTF2_LocationRef 	location,
                            OTF2_TimeStamp 	time,
                            void*  userData,
                            OTF2_AttributeList *attributeList,
                            OTF2_StringRef programName,
                            uint32_t numberOfArguments,
                            const OTF2_StringRef *programArguments )
{
    global_trace_loader->last_read_timestamp = time;


    if ( global_trace_loader->calls.find(location) ==  global_trace_loader->calls.end())
    {
         global_trace_loader->calls[location].push( new TraceEvent(  location )); // add an _root element
    }

    std::string& name = global_trace_loader->string_defs[programName];
    std::string::size_type found = name.rfind("/");
    std::string region_name = "";
    if (found!=std::string::npos)
       region_name = name.substr(++found);
    uint64_t region = OTF2_UNDEFINED_REGION;
    for ( std::vector<RegionClass>::const_iterator riter = global_trace_loader->regions.begin(); riter != global_trace_loader->regions.end(); ++riter,++region)
    {
        region= (region == OTF2_UNDEFINED_REGION)?0:region;
        if (riter->name.compare( region_name) == 0)
            break;
    }
    TraceEvent * _event = TraceLoader::createTraceEvent(OTF2_UNDEFINED_REGION, location ) ;

    _event->set_start( time - global_trace_loader->global_offset  );
    _event->set_region( region );
    _event->set_paradigm( global_trace_loader->regions[ region].paradigm);
    _event->set_region_role(global_trace_loader->regions[ region].regionRole );
    _event->set_level( global_trace_loader->calls[location].size() );

    TraceEvent * caller =global_trace_loader->calls[location].top();
    _event->set_parent( caller );
    global_trace_loader->max_levels[location] = std::max(global_trace_loader->max_levels[location], _event->get_level() );
    uint32_t parent_callpath = (_event->get_parent() != NULL)? _event->get_parent()->get_callpath():OTF2_UNDEFINED_CALLPATH;
    uint32_t current_callpath = find_callpath( caller, parent_callpath, region);

    _event->set_callpath( current_callpath );
    _event->set_cube_callpath( bladeservices::get_cube_callpath( current_callpath, global_trace_loader->cube ));
    _event->set_cube_location( bladeservices::get_cube_location( location,  global_trace_loader->cube ));

    global_trace_loader->current_metric_values.clear();

    global_trace_loader->calls[location].push( _event);
    return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_ProgramEnd( OTF2_LocationRef 	location,
                        OTF2_TimeStamp 	time,
                        void *userData,
                        OTF2_AttributeList *attributeList,
                        int64_t exitStatus)
{
    TraceEvent * _event = global_trace_loader->calls[location].top();
    global_trace_loader->calls[location].pop();

    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;

    _event->set_end( time - global_trace_loader->global_offset  );
    global_trace_loader->last_read_timestamp = time;
    return OTF2_CALLBACK_SUCCESS;
}







static OTF2_CallbackCode
TraceLoader_Enter(  OTF2_LocationRef    location,
                    OTF2_TimeStamp      time,
                    void*               userData,
                    OTF2_AttributeList* attributes,
                    OTF2_RegionRef      region )
{
    global_trace_loader->last_read_timestamp = time;


    if ( global_trace_loader->calls.find(location) ==  global_trace_loader->calls.end())
    {
         global_trace_loader->calls[location].push( new TraceEvent(  location )); // add an _root element
    }



    TraceEvent * _event = TraceLoader::createTraceEvent(region, location ) ;





    _event->set_start( time - global_trace_loader->global_offset  );
    _event->set_region( region );
    _event->set_paradigm( global_trace_loader->regions[ region].paradigm);
    _event->set_region_role(global_trace_loader->regions[ region].regionRole );
    _event->set_level( global_trace_loader->calls[location].size() );

    TraceEvent * caller =global_trace_loader->calls[location].top();
    _event->set_parent( caller );

    if (global_trace_loader->trace_load_enter_filter->apply( _event ) != FILTER_OK)
    {
        delete _event;
        global_trace_loader->calls[location].push(  global_trace_loader->dummy_event);
            return OTF2_CALLBACK_SUCCESS;
    }

    global_trace_loader->max_levels[location] = std::max(global_trace_loader->max_levels[location], _event->get_level() );
    uint32_t parent_callpath = (_event->get_parent() != NULL)? _event->get_parent()->get_callpath():OTF2_UNDEFINED_CALLPATH;
    uint32_t current_callpath = find_callpath( caller, parent_callpath, region);

    _event->set_callpath( current_callpath );
    _event->set_cube_callpath( bladeservices::get_cube_callpath( current_callpath, global_trace_loader->cube ));
    _event->set_cube_location( bladeservices::get_cube_location( location,  global_trace_loader->cube ));

    _event->set_start_metrics( global_trace_loader->current_metric_values);
    global_trace_loader->current_metric_values.clear();




    if (global_trace_loader->grouping)
    {
        OmpParallelGroup* _mg = global_trace_loader->omp_parallel_groups->add_event( _event, bladeservices::master_id(location) );
        if (_mg != NULL)
            _event->set_events_group(_mg);
    }
    global_trace_loader->calls[location].push( _event);


    return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_Leave(  OTF2_LocationRef    location,
                    OTF2_TimeStamp      time,
                    void*               userData,
                    OTF2_AttributeList* attributes,
                    OTF2_RegionRef      region )
{
    TraceEvent * _event = global_trace_loader->calls[location].top();
    global_trace_loader->calls[location].pop();

    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;

    _event->set_end( time- global_trace_loader->global_offset  );
    _event->set_end_metrics( global_trace_loader->current_metric_values);
    global_trace_loader->current_metric_values.clear();

    global_trace_loader->last_read_timestamp = time;

    if (global_trace_loader->trace_load_leave_filter->apply( _event ) != FILTER_OK)
    {
        // remove this event and move children to parent
        _event->reparent_children();
        if (global_trace_loader->grouping)
        {
            global_trace_loader->mpi_p2p_groups->find_and_delete( _event );
            global_trace_loader->mpi_collective_groups->find_and_delete( _event );
            global_trace_loader->omp_parallel_groups->find_and_delete( _event );
        }
        delete _event;
    }
    return OTF2_CALLBACK_SUCCESS;
}



static OTF2_CallbackCode
TraceLoader_MetricCallback( OTF2_LocationRef locationID,
                            OTF2_TimeStamp time,
                            void *userData,
                            OTF2_AttributeList *attributeList,
                            OTF2_MetricRef metric,
                            uint8_t numberOfMetrics,
                            const OTF2_Type *typeIDs,
                            const OTF2_MetricValue *metricValues)
{
    for (size_t i = 0; i<numberOfMetrics; i++)
    {
        OTF2_MetricMemberRef metric_index =  global_trace_loader->trace->get_metric_index( metric, i);
        if ( metric_index >= global_trace_loader->current_metric_values.size() )
            global_trace_loader->current_metric_values.resize( metric_index+1);
        global_trace_loader->current_metric_values[metric_index] = metricValues[i].unsigned_int;
    }
    return OTF2_CALLBACK_SUCCESS;
}





static OTF2_CallbackCode
TraceLoader_GlobDefLocation(   void*                 userData,
                                        OTF2_LocationRef      location,
                                        OTF2_StringRef        name,
                                        OTF2_LocationType     locationType,
                                        uint64_t              numberOfEvents,
                                        OTF2_LocationGroupRef locationGroup )
{
    global_trace_loader->locations = (trace_vector* )userData;

    if ( global_trace_loader->locations->size == global_trace_loader->locations->capacity )
    {
        return OTF2_CALLBACK_INTERRUPT;
    }

    global_trace_loader->locations->members[ global_trace_loader->locations->size ] = location;
//     global_trace_loader->locations->numberOfEvents[ global_trace_loader->locations->size ] = numberOfEvents;
    global_trace_loader->locations->size++;

    uint64_t _master_id = bladeservices::master_id(location);
    uint64_t _thread_id = bladeservices::thread_id(location);

    if (_thread_id != 0) // non master openmpthreads....
    {
        global_trace_loader->forks[ _master_id ][ location ] = 1;
    }
    std::stringstream ss;
    ss << _thread_id;
    std::string _thread_name = ss.str();


    LocationClass _loc;
    _loc.self = location;
    if (_thread_id == 0)
    {
        _loc.name = global_trace_loader->lg_classes[locationGroup].name;
        global_trace_loader->number_of_ranks++;
    }
    else
        _loc.name = (global_trace_loader->string_defs[name].compare("")==0)?"Thread "+_thread_name:global_trace_loader->string_defs[name];
    _loc.numberOfEvents = numberOfEvents;

    global_trace_loader->locations_classes[location] = _loc;
    global_trace_loader->lg_classes[locationGroup].locations.push_back(location);
    return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_GlobDefLocationGroup(   void* userData,
                                    OTF2_LocationGroupRef  self,
                                    OTF2_StringRef         name,
                                    OTF2_LocationGroupType locationGroupType,
                                    OTF2_SystemTreeNodeRef systemTreeParent,
                                    OTF2_LocationGroupRef  creatingLocationGroup )
{

    LocationGroupClass _lg;
    _lg.self = self;
    _lg.name = global_trace_loader->string_defs[name];
    global_trace_loader->lg_classes[self] = _lg;
    global_trace_loader->stn_classes[systemTreeParent].groups.push_back(self);
    return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_GlobDefSystemTreeNode(   void *userData,
                                     OTF2_SystemTreeNodeRef self,
                                     OTF2_StringRef name,
                                     OTF2_StringRef className,
                                     OTF2_SystemTreeNodeRef parent )
{

    SystemTreeNodeClass _stn;
    _stn.self = self;
    _stn.name = global_trace_loader->string_defs[name];
    global_trace_loader->stn_classes[self] = _stn;

    if (parent == OTF2_UNDEFINED_SYSTEM_TREE_NODE )
    {
        global_trace_loader->root_stn_classes.push_back(self);
    }else
    {
        global_trace_loader->stn_classes[parent].children.push_back(self);
    }


    return OTF2_CALLBACK_SUCCESS;
}







// ---------- callbacks for groups of events ----------------

static OTF2_CallbackCode
TraceLoader_MpiCollectiveBegin(  OTF2_LocationRef locationID,
                                 OTF2_TimeStamp time,
                                 void *userData,
                                 OTF2_AttributeList *attributeList
                              )
{
    TraceEvent * _event = global_trace_loader->calls[locationID].top();
    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;

    if (global_trace_loader->grouping)
    {
        MpiCollectiveGroup* _mg = global_trace_loader->mpi_collective_groups->add_event( _event, bladeservices::master_id(locationID));
        _event->set_events_group(_mg);
    }
    return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_MpiCollectiveEnd(  OTF2_LocationRef locationID,
                              OTF2_TimeStamp time,
                              void *userData,
                              OTF2_AttributeList *attributeList,
                              OTF2_CollectiveOp collectiveOp,
                              OTF2_CommRef communicator,
                              uint32_t root,
                              uint64_t sizeSent,
                              uint64_t sizeReceived
                           )
{
    TraceEvent * _event = global_trace_loader->calls[locationID].top();
    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;

    BcastTraceEvent * _as_bcast = dynamic_cast<BcastTraceEvent*>(_event);
    ReduceTraceEvent * _as_reduce = dynamic_cast<ReduceTraceEvent*>(_event);
    ScanTraceEvent * _as_scan = dynamic_cast<ScanTraceEvent*>(_event);
    GatherTraceEvent * _as_gather = dynamic_cast<GatherTraceEvent*>(_event);

    if (_as_bcast != NULL || _as_reduce != NULL || _as_gather !=NULL)
    {
        if (root == bladeservices::master_id(locationID))
            _event->set_group_leader(true);
    }

        return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_MpiSend(  OTF2_LocationRef locationID,
                      OTF2_TimeStamp time,
                      void *userData,
                      OTF2_AttributeList *attributeList,
                      uint32_t receiver,
                      OTF2_CommRef communicator,
                      uint32_t msgTag,
                      uint64_t msgLength
                              )
{
    TraceEvent * _event = global_trace_loader->calls[locationID].top();
    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;


    SendP2PTraceEvent * _as_send = dynamic_cast<SendP2PTraceEvent*>(_event);
    if (_as_send == NULL) // if not send -> ignore
        return OTF2_CALLBACK_SUCCESS;

    _as_send->init( msgTag, receiver, communicator );

    if (global_trace_loader->grouping)
    {
        global_trace_loader->mpi_p2p_groups->add_event( _as_send, bladeservices::master_id(locationID));
    }
    return OTF2_CALLBACK_SUCCESS;
}

static OTF2_CallbackCode
TraceLoader_MpiRecv( OTF2_LocationRef locationID,
                     OTF2_TimeStamp time,
                     void *userData,
                     OTF2_AttributeList *attributeList,
                     uint32_t sender,
                     OTF2_CommRef communicator,
                     uint32_t msgTag,
                     uint64_t msgLength
                           )
{

    TraceEvent * _event = global_trace_loader->calls[locationID].top();
    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;


    RecvP2PTraceEvent * _as_recv = dynamic_cast<RecvP2PTraceEvent*>(_event);
    if (_as_recv == NULL) // if not recv -> ignore
        return OTF2_CALLBACK_SUCCESS;
    _as_recv->init( msgTag, sender, communicator );

    if (global_trace_loader->grouping)
    {
        global_trace_loader->mpi_p2p_groups->add_event( _as_recv, bladeservices::master_id(locationID));
    }
    return OTF2_CALLBACK_SUCCESS;
}




static OTF2_CallbackCode
TraceLoader_MpiIsend( OTF2_LocationRef locationID,
                      OTF2_TimeStamp time,
                      void *userData,
                      OTF2_AttributeList *attributeList,
                      uint32_t receiver,
                      OTF2_CommRef communicator,
                      uint32_t msgTag,
                      uint64_t msgLength,
                      uint64_t requestID
                           )
{
    TraceEvent * _event = global_trace_loader->calls[locationID].top();
    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;


    ISendP2PTraceEvent * _as_isend = dynamic_cast<ISendP2PTraceEvent*>(_event);
    if (_as_isend == NULL) // if not isend -> ignore
        return OTF2_CALLBACK_SUCCESS;

    _as_isend->init( msgTag, receiver, communicator, requestID );
    if (global_trace_loader->grouping)
    {
        global_trace_loader->mpi_p2p_groups->add_incomplete_event( _as_isend, bladeservices::master_id(locationID));
    }
    return OTF2_CALLBACK_SUCCESS;
}


static OTF2_CallbackCode
TraceLoader_MpiIsendComplete( OTF2_LocationRef locationID,
                              OTF2_TimeStamp time,
                              void *userData,
                              OTF2_AttributeList *attributeList,
                              uint64_t requestID
                           )
{
    TraceEvent * _event = global_trace_loader->calls[locationID].top();
    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;


    WaitTraceEvent * _as_wait = dynamic_cast<WaitTraceEvent*>(_event);
    WaitallTraceEvent * _as_waitall = dynamic_cast<WaitallTraceEvent*>(_event);

    if (_as_wait != NULL)
    {
        _as_wait->set_request( requestID );
    }else
    if (_as_waitall != NULL)
    {
        _as_waitall->set_request( requestID );
    }else // exit if none of them
          return OTF2_CALLBACK_SUCCESS;
    if (global_trace_loader->grouping)
    {
        global_trace_loader->mpi_p2p_groups->add_event( _event, bladeservices::master_id(locationID));
        global_trace_loader->mpi_p2p_groups->complete_isend_event( _event, bladeservices::master_id(locationID));
    }
    return OTF2_CALLBACK_SUCCESS;
}


static OTF2_CallbackCode
TraceLoader_MpiIrecvRequest(OTF2_LocationRef locationID,
                            OTF2_TimeStamp time,
                            void *userData,
                            OTF2_AttributeList *attributeList,
                            uint64_t requestID
                           )
{
    TraceEvent * _event = global_trace_loader->calls[locationID].top();
    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;


    IRecvP2PTraceEvent * _as_irecv = dynamic_cast<IRecvP2PTraceEvent*>(_event);
    if (_as_irecv == NULL) // if not _as_irecv -> ignore
        return OTF2_CALLBACK_SUCCESS;
    _as_irecv->set_request( requestID );
    if (global_trace_loader->grouping)
    {
        global_trace_loader->mpi_p2p_groups->add_incomplete_event( _as_irecv, bladeservices::master_id(locationID));
    }
    return OTF2_CALLBACK_SUCCESS;
}


static OTF2_CallbackCode
TraceLoader_MpiIrecv( OTF2_LocationRef locationID,
                      OTF2_TimeStamp time,
                      void *userData,
                      OTF2_AttributeList *attributeList,
                      uint32_t sender,
                      OTF2_CommRef communicator,
                      uint32_t msgTag,
                      uint64_t msgLength,
                      uint64_t requestID
                           )
{
    TraceEvent * _event = global_trace_loader->calls[locationID].top();
    if (_event == global_trace_loader->dummy_event)
        return OTF2_CALLBACK_SUCCESS;


    WaitTraceEvent * _as_wait = dynamic_cast<WaitTraceEvent*>(_event);
    WaitallTraceEvent * _as_waitall = dynamic_cast<WaitallTraceEvent*>(_event);

    if (_as_wait != NULL)
    {
        _as_wait->init(  msgTag, sender, communicator, requestID );
    }else
    if (_as_waitall != NULL)
    {
        _as_waitall->init(  msgTag, sender, communicator, requestID );
    }else // ignore no any of them
        return OTF2_CALLBACK_SUCCESS;
    if (global_trace_loader->grouping)
    {
        global_trace_loader->mpi_p2p_groups->add_event( _event, bladeservices::master_id(locationID));
        global_trace_loader->mpi_p2p_groups->complete_irecv_event( _event, bladeservices::master_id(locationID));
    }
    return OTF2_CALLBACK_SUCCESS;
}




// ------------------- static methods ----------------


TraceEvent *
TraceLoader::createTraceEvent(uint64_t region,  uint64_t location)
{
    if (region != OTF2_UNDEFINED_REGION)
    {
        if (
            (global_trace_loader->regions[ region].name.compare("MPI_Send")==0) ||
            (global_trace_loader->regions[ region].name.compare("MPI_Ssend")==0)||
            (global_trace_loader->regions[ region].name.compare("MPI_Rsend")==0)
            )
        {
            return  new SendP2PTraceEvent(  location);
        }
        else if  (global_trace_loader->regions[ region].name.compare("MPI_Recv")==0)
        {
            return new RecvP2PTraceEvent(  location);
        }
        else if ( (global_trace_loader->regions[ region].name.compare("MPI_Isend")==0) ||
                (global_trace_loader->regions[ region].name.compare("MPI_Issend")==0)||
                (global_trace_loader->regions[ region].name.compare("MPI_Irsend")==0)
                )
        {
            return  new ISendP2PTraceEvent(  location);
        }
        else if  (global_trace_loader->regions[ region].name.compare("MPI_Irecv")==0)
        {
            return new IRecvP2PTraceEvent(  location);
        }
        else if (global_trace_loader->regions[ region].name.compare("MPI_Bcast")==0)
        {
            return new BcastTraceEvent(  location);
        }
        else if  (global_trace_loader->regions[ region].name.compare("MPI_Reduce")==0)
        {
            return new ReduceTraceEvent(  location);
        }
        else  if (global_trace_loader->regions[ region].name.compare("MPI_Scan")==0)
        {
            return new ScanTraceEvent(  location);
        }
        else if  (global_trace_loader->regions[ region].name.compare("MPI_Gather")==0)
        {
            return new GatherTraceEvent(  location);
        }
        else if  (global_trace_loader->regions[ region].name.compare("MPI_Wait")==0)
        {
            return new WaitTraceEvent(  location);
        }
        else if  (global_trace_loader->regions[ region].name.compare("MPI_Waitall")==0)
        {
            return new WaitallTraceEvent(  location);
        }
    }

    // i case of doublt -> create general trace
    return new TraceEvent( location );
}


// ---------------------------------------------------






TraceLoader::TraceLoader( std::string _path,cube::Cube * _cube ):cube(_cube)
{
    grouping = true;
    dummy_event = NULL;
    number_of_ranks = 0;
    callpath_filter_active = false;
    selection_filter_active = false;
    trace_path = _path;
    global_trace_loader = this;
    start_timestamp = std::numeric_limits<uint64_t>::max();
    end_timestamp = std::numeric_limits<uint64_t>::min();
    trace = new Trace( );
    trace->set_trace_path(trace_path );


    reader = OTF2_Reader_Open( trace_path.c_str() );

    OTF2_Reader_SetSerialCollectiveCallbacks( reader );

    OTF2_Reader_GetNumberOfLocations( reader,
                                      &number_of_locations );
    locations = (trace_vector*)malloc( sizeof( *locations )
                                       + number_of_locations
                                       * sizeof( *locations->members )
                                     );
    locations->capacity = number_of_locations;
    locations->size     = 0;

    //
    global_def_reader = OTF2_Reader_GetGlobalDefReader( reader );
//
    global_def_callbacks = OTF2_GlobalDefReaderCallbacks_New();
    OTF2_GlobalDefReaderCallbacks_SetSystemTreeNodeCallback( global_def_callbacks,
                                                       &TraceLoader_GlobDefSystemTreeNode );
    OTF2_GlobalDefReaderCallbacks_SetLocationGroupCallback( global_def_callbacks,
                                                       &TraceLoader_GlobDefLocationGroup );
    OTF2_GlobalDefReaderCallbacks_SetLocationCallback( global_def_callbacks,
                                                       &TraceLoader_GlobDefLocation );

    OTF2_GlobalDefReaderCallbacks_SetClockPropertiesCallback( global_def_callbacks,
                                                       &TraceLoader_ClockProperties );
    OTF2_GlobalDefReaderCallbacks_SetStringCallback( global_def_callbacks,
                                                     &TraceLoader_StringDefinition );
    OTF2_GlobalDefReaderCallbacks_SetRegionCallback( global_def_callbacks,
                                                     &TraceLoader_RegionDefinition);
    OTF2_GlobalDefReaderCallbacks_SetCallpathCallback( global_def_callbacks,
                                                       &TraceLoader_CallPathDefinition);

    OTF2_GlobalDefReaderCallbacks_SetMetricMemberCallback( global_def_callbacks,
                                                       &TraceLoader_MetricMemberDefinition);
    OTF2_GlobalDefReaderCallbacks_SetMetricClassCallback ( global_def_callbacks,
                                                       &TraceLoader_MetricClassDefinition);
    OTF2_GlobalDefReaderCallbacks_SetMetricInstanceCallback( global_def_callbacks,
                                                       &TraceLoader_MetricInstanceDefinition);
//     OTF2_GlobalEvtReaderCallbacks_SetMetricCallback ( 	 global_def_callbacks,
//                                                        &TraceLoader_MetricDefinition)


    OTF2_Reader_RegisterGlobalDefCallbacks( reader,
                                            global_def_reader,
                                            global_def_callbacks,
                                            locations );

    loadDefinitions();

}

TraceLoader::~TraceLoader(  )
{
    OTF2_GlobalEvtReaderCallbacks_Delete( event_callbacks );


    OTF2_Reader_CloseGlobalEvtReader( reader, global_evt_reader );

    if ( successful_open_def_files )
    {
        OTF2_GlobalDefReaderCallbacks_Delete( global_def_callbacks );
        OTF2_Reader_CloseDefFiles( reader );
    }


    OTF2_Reader_CloseEvtFiles( reader );

    OTF2_Reader_Close( reader );


    free(locations);

}


Trace *
TraceLoader::getTrace()
{
    return trace;
}


void
TraceLoader::loadDefinitions( )
{




uint64_t definitions_read = 0;
OTF2_Reader_ReadAllGlobalDefinitions(   reader,
                                            global_def_reader,
                                            &definitions_read );


    for ( size_t i = 0; i < locations->size; i++ )
    {
        OTF2_Reader_SelectLocation( reader, locations->members[ i ] );
    }

    successful_open_def_files =
        OTF2_Reader_OpenDefFiles( reader ) == OTF2_SUCCESS;
    OTF2_Reader_OpenEvtFiles( reader );
    for ( size_t i = 0; i < locations->size; i++ )
    {
        if ( successful_open_def_files )
        {
            OTF2_DefReader* def_reader =
                OTF2_Reader_GetDefReader( reader, locations->members[ i ] );
            if ( def_reader )
            {
                uint64_t def_reads = 0;
                OTF2_Reader_ReadAllLocalDefinitions( reader,
                                                     def_reader,
                                                     &def_reads );
                OTF2_Reader_CloseDefReader( reader, def_reader );
            }
            OTF2_Reader_GetEvtReader( reader, locations->members[i] );
        }
    }





    trace->init( start_timestamp, end_timestamp,  timer_resolution, global_offset, trace_length, regions, calltree);
    trace->set_system_tree( root_stn_classes, stn_classes, lg_classes, locations_classes);
    trace->set_callpath_filter_active( callpath_filter_active);
    trace->set_selection_filter_active( selection_filter_active);
    trace->set_metrics( metric_members, metric_classes, metric_instances);
    mpi_collective_groups = new MpiCollectiveGroupsCollection( number_of_ranks );
    mpi_p2p_groups = new MpiP2PGroupsCollection( number_of_ranks );
    omp_parallel_groups = new OmpParallelGroupsCollection(number_of_ranks);

}



void
TraceLoader::loadEvents( TraceLoadEnterFilter * _tlf_enter, TraceLoadLeaveFilter * _tlf_leave, bool _grouping)
{
    grouping = _grouping;
    trace_load_enter_filter = _tlf_enter;
    trace_load_leave_filter = _tlf_leave;
    progress_message = "Reading events...";
    progress_step =  bladeservices::TraceReading;
    global_evt_reader = OTF2_Reader_GetGlobalEvtReader( reader );
    event_callbacks = OTF2_GlobalEvtReaderCallbacks_New();

    OTF2_GlobalEvtReaderCallbacks_SetProgramBeginCallback ( event_callbacks,
                                                            &TraceLoader_ProgramBegin );
    OTF2_GlobalEvtReaderCallbacks_SetProgramEndCallback ( event_callbacks,
                                                            &TraceLoader_ProgramEnd );

    OTF2_GlobalEvtReaderCallbacks_SetEnterCallback( event_callbacks,
                                                    &TraceLoader_Enter );
    OTF2_GlobalEvtReaderCallbacks_SetLeaveCallback( event_callbacks,
                                                    &TraceLoader_Leave );
    OTF2_GlobalEvtReaderCallbacks_SetOmpForkCallback ( event_callbacks,
                                                       &TraceLoader_OmpFork);
    OTF2_GlobalEvtReaderCallbacks_SetOmpJoinCallback ( event_callbacks,
                                                       &TraceLoader_OmpJoin);
    OTF2_GlobalEvtReaderCallbacks_SetThreadForkCallback ( event_callbacks,
                                                       &TraceLoader_ThreadFork);
    OTF2_GlobalEvtReaderCallbacks_SetThreadJoinCallback ( event_callbacks,
                                                       &TraceLoader_ThreadJoin);
     OTF2_GlobalEvtReaderCallbacks_SetThreadCreateCallback ( event_callbacks,
                                                       &TraceLoader_ThreadBegin);
     OTF2_GlobalEvtReaderCallbacks_SetThreadBeginCallback ( event_callbacks,
                                                       &TraceLoader_ThreadBegin);
    OTF2_GlobalEvtReaderCallbacks_SetThreadEndCallback ( event_callbacks,
                                                       &TraceLoader_ThreadEnd);
    OTF2_GlobalEvtReaderCallbacks_SetMpiCollectiveBeginCallback ( event_callbacks,
                                                                &TraceLoader_MpiCollectiveBegin);

    OTF2_GlobalEvtReaderCallbacks_SetMpiCollectiveEndCallback ( event_callbacks,
                                                       &TraceLoader_MpiCollectiveEnd);

     OTF2_GlobalEvtReaderCallbacks_SetMpiSendCallback( event_callbacks,
                                                        &TraceLoader_MpiSend);
     OTF2_GlobalEvtReaderCallbacks_SetMpiRecvCallback( event_callbacks,
                                                        &TraceLoader_MpiRecv);

    OTF2_GlobalEvtReaderCallbacks_SetMpiIsendCallback ( event_callbacks,
                                                        &TraceLoader_MpiIsend);
    OTF2_GlobalEvtReaderCallbacks_SetMpiIsendCompleteCallback (event_callbacks,
                                                        &TraceLoader_MpiIsendComplete);
    OTF2_GlobalEvtReaderCallbacks_SetMpiIrecvRequestCallback ( event_callbacks,
                                                        &TraceLoader_MpiIrecvRequest);
    OTF2_GlobalEvtReaderCallbacks_SetMpiIrecvCallback(event_callbacks,
                                                        &TraceLoader_MpiIrecv );






    OTF2_GlobalEvtReaderCallbacks_SetMetricCallback ( event_callbacks,
                                                       &TraceLoader_MetricCallback);


    OTF2_Reader_RegisterGlobalEvtCallbacks( reader,
                                            global_evt_reader,
                                            event_callbacks,
                                            NULL );

    uint64_t events_read = 0;
    OTF2_Reader_ReadAllGlobalEvents( reader,
                                     global_evt_reader,
                                     &events_read );


    std::map<uint64_t, TraceEvent*> _trace_entries;
    for (std::map<uint64_t, std::stack<TraceEvent*> >::iterator iter=calls.begin(); iter!= calls.end(); ++iter)
    {
        TraceEvent * _entry = iter->second.top();
        _trace_entries[iter->first] = _entry;
    }

    if (grouping)
    {
        progress_step =  bladeservices::P2PMatching;
        global_trace_loader->mpi_p2p_groups->match_events();
        progress_step =  bladeservices::P2PGrouping;
        global_trace_loader->mpi_p2p_groups->group_events();
    }






    trace->set_events( _trace_entries );
    trace->set_max_levels(max_levels);
    trace->set_mpi_collective_groups_collection( mpi_collective_groups );
    trace->set_mpi_p2p_groups_collection( mpi_p2p_groups );
    trace->set_omp_parallel_groups_collection( omp_parallel_groups );
}






double
TraceLoader::get_progress()
{
    switch (progress_step)
    {
        case  bladeservices::TraceReading:
            return (double)(last_read_timestamp - global_offset) / (double)( end_timestamp );
        case  bladeservices::P2PMatching:
            return global_trace_loader->mpi_p2p_groups->get_progress( bladeservices::P2PMatching);
        case  bladeservices::P2PGrouping:
            return global_trace_loader->mpi_p2p_groups->get_progress( bladeservices::P2PGrouping);
        default:
            return 1.;
    }
}


std::string&
TraceLoader::get_progress_message()
{
    switch (progress_step)
    {
        case  bladeservices::TraceReading:
            return progress_message;
        case  bladeservices::P2PMatching:
        case  bladeservices::P2PGrouping:
            return global_trace_loader->mpi_p2p_groups->get_progress_message();
        default:
            return progress_message;
    }
}
