/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "config.h"
#include <iostream>
#include <fstream>

#include <QLabel>
#include <QGraphicsView>
#include <QGraphicsScene>
#include <QRectF>
#include <QVBoxLayout>
#include <QScrollArea>
#include <QAction>
#include <QFile>
#include <QFileDialog>
#include <QStringList>
#include <QString>
#include <QGroupBox>
#include <QMessageBox>

#include "CubeIdIndexMap.h"
#include "ScorePionPluginWidget.h"
#include "ScorePionFilterSetWidget.h"
#include "ScorePionToolBar.h"

#include "ExcludeFileFilterRulesSet.h"
#include "ExcludeRegionFilterRulesSet.h"
#include "IncludeFileFilterRulesSet.h"
#include "IncludeRegionFilterRulesSet.h"

#include "ScorePionFiltersFactory.h"
#include "scorep/SCOREP_Score_Profile.hpp"
#include "scorep/SCOREP_Score_Estimator.hpp"

using namespace scorepion_plugin::gui;
using namespace scorepion_plugin::filters;

extern  cubepluginapi::PluginServices* scorepion_service;

ScorePionPluginWidget::ScorePionPluginWidget(cube::CubeProxy* _cube,  ScorePionToolBar* toolbar, scorepion_plugin::filters::ScorePionFiltersFactory * _filter_factory, QWidget * parent):QWidget(parent),toolBar(toolbar), cube(_cube),filter_factory(_filter_factory)
{
    reduced_trace_size_metric =  NULL;
    total_trace_size_metric =  NULL;

    cubepl_memory = NULL;
    variable_adress = 0;
    region_bytes_adress = 0;

    toolBar->setScorePionPluginWidget(this);

    filters = new scorepion_plugin::filters::FiltersChain();

    QVBoxLayout * top_layout = new QVBoxLayout();
    this->setLayout( top_layout );
    QGroupBox * metrics_frame = new QGroupBox( "Measurement" );
    QGridLayout * grid_layout = new QGridLayout ( );
    metrics_frame->setLayout( grid_layout);
    top_layout->addWidget( metrics_frame );

    QLabel * full_trace_size_label = new QLabel ("Full trace size ");
    grid_layout->addWidget( full_trace_size_label , 0, 0);

    full_trace_size = new QLabel ("---");
    grid_layout->addWidget( full_trace_size , 0, 1);

    QLabel * reduced_trace_size_label = new QLabel ("Reduced trace size ");
    grid_layout->addWidget( reduced_trace_size_label , 1, 0);

    reduced_trace_size = new QLabel ("---");
    grid_layout->addWidget( reduced_trace_size , 1, 1);

    QLabel * scorep_total_memory_label = new QLabel ("SCOREP_TOTAL_MEMORY");
    grid_layout->addWidget( scorep_total_memory_label , 2, 0);
    scorep_total_memory = new QLabel ("---");
    grid_layout->addWidget( scorep_total_memory , 2, 1);

    cube::Metric * _met = cube->getMetric("time");
    cube::IdIndexMap                  metric_id_indices;
    cube::value_container       inclusive_values;
    cube::value_container       exclusive_values;
    cube::list_of_cnodes       cnode_selections;
    const std::vector< cube::Cnode* > &croots = cube->getRootCnodes();
    for( std::vector< cube::Cnode* >::const_iterator citer = croots.begin(); citer!= croots.end(); ++citer )
    {
        cube::list_of_cnodes cnodes;
        cube::cnode_pair     pair;
        pair.first  = *citer;
        pair.second = cube::CUBE_CALCULATE_INCLUSIVE;
        cnode_selections.push_back( pair );
    }
    cube::list_of_sysresources sysres_selection;
    cube->getMetricSubtreeValues(   cnode_selections,
                                    sysres_selection,
                                    *_met,
                                    static_cast<size_t>(0),
                                    metric_id_indices,
                                    &inclusive_values,
                                    &exclusive_values );
    total_runtime = (inclusive_values[0])?inclusive_values[0]->getDouble():0.;



    QGroupBox * filters_frame = new QGroupBox( "Filter rules" );
    main_layout    = new QVBoxLayout();
    filters_frame->setLayout( main_layout);
    top_layout->addWidget( filters_frame );


    std::list<FilterRulesSet *> &f_sets = filter_factory->getFilterSets();

    if (f_sets.size() != 0)
    {
        for ( std::list<FilterRulesSet *>::iterator iter = f_sets.begin(); iter != f_sets.end(); ++iter)
        {
            FilterRulesSet * _f_set = *iter;
            addFilterRule( _f_set );
        }
    }
    else
    {
    	addDefaultFilterRule( );
    }

    defineTreeItemMarker();

    updateMarkers();

}


ScorePionPluginWidget::~ScorePionPluginWidget()
{
    delete filters;
}



void
ScorePionPluginWidget::writeFilterFile( )
{
    QString fileName = QFileDialog::getSaveFileName(this, tr("Save Score-P's Filter file"),
                           "./new_filter.filt",
                           tr("ScoreP Filter Files (*.filt );;Intel Filter File (*.iflt)"));

	QFileInfo fi (fileName);
	bool intel = (fi.suffix().compare("iflt") == 0);

	std::fstream fout;
	fout.open(fileName.toUtf8().data(), std::ios_base::out );
	fout << "#Filter File created by ScorePion Plugin" << std::endl;
	fout << "#Plugin autor: Pavel Saviankou" << std::endl;
	fout << "#Url: Scalasca.org" << std::endl;
	for ( std::list<ScorePionFilterSetWidget *>::const_iterator iter = filter_sets.begin(); iter != filter_sets.end(); ++iter)
	{
	FilterRulesSet * _f_set = (*iter)->getFilterRules();
		if ( intel )
		{
			_f_set->writeIntel( fout );
		}
		else
		{
			_f_set->write( fout );
		}

	}
        fout.close();
    	QString variables;
    	variables+= tr("Set in your job script:\n\n");
    	variables+= tr("SCOREP_TOTAL_MEMORY=") + scorep_total_memory->text() + "\n";
    	if (!intel)
    	{
    		variables+= tr("SCOREP_FILTERING_FILE=") + fileName + "\n";
	}
	else
	{
		variables += "\n";
		variables+=tr("Compile your project using Intel Compiler suite with command line option: \n");
		variables+=tr("'-tcollect-filter ") + fileName + "'";
	}
	QMessageBox * msgBox = new QMessageBox(this);
	msgBox->setText(variables);
	msgBox->exec();
}



void
ScorePionPluginWidget::loadFilterFile( )
{
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Score-P's Filter file"),
                           "./filter.flt",
                           tr("ScoreP Filter Files (*.filt );;All Files (*)"));

      filter_factory->readFilterFile(fileName);
        std::list<FilterRulesSet *> &f_sets = filter_factory->getFilterSets();

        if (f_sets.size() != 0)
        {
            for ( std::list<FilterRulesSet *>::iterator iter = f_sets.begin(); iter != f_sets.end(); ++iter)
            {
                FilterRulesSet * _f_set = *iter;
                addFilterRule( _f_set );
            }
        }
    applyFilters();
}


/** defines a number of tree items markers, which are used to mark tree items with different background colour and
    icons */
void
ScorePionPluginWidget::defineTreeItemMarker()
{
        QList<QPixmap> icons;
        icons.append( QPixmap( ":scorepionicons/add_to_filter.png" ) );
        markerList.append( scorepion_service->getTreeItemMarker( "Include in measurement", icons, false ) );
        icons.clear();
        icons.append( QPixmap( ":scorepionicons/remove_from_filter.png" ) );
        markerList.append( scorepion_service->getTreeItemMarker( "Exclude from measurement", icons, true ) );
        icons.clear();
        icons.append( QPixmap( ":scorepionicons/conflict_filter.png" ) );
        markerList.append( scorepion_service->getTreeItemMarker( "Conflicting rules", icons, false ) );
        icons.clear();
        icons.append( QPixmap( ":scorepionicons/icon.png" ) );
        markerList.append( scorepion_service->getTreeItemMarker( "Filter", icons, false ) );
}


/** defines a number of tree items markers, which are used to mark tree items with different background colour and
    icons */
void
ScorePionPluginWidget::updateMarkers()
{
    scorepion_service->removeMarker();
    const QList<cubepluginapi::TreeItem*> _all_callpath_items = scorepion_service->getTreeItems( cubepluginapi::DEFAULTCALLTREE );
    foreach( cubepluginapi::TreeItem * _item, _all_callpath_items )
    {

        cube::Vertex * _vertex = _item->getCubeObject();

        if (_vertex != NULL)
        {
            double index = getRegionId(_vertex );
            if (cubepl_memory!=NULL &&  index != -1)
                cubepl_memory->put(variable_adress, index, 1, 0, cube::CUBEPL_GLOBAL_VARIABLE );

            scorepion_plugin::filters::FilterResult result = filters->apply( _vertex );

            if ( result == INCLUDE )
            {
                if (cubepl_memory != NULL)
                    cubepl_memory->put(variable_adress, index, 1., 0, cube::CUBEPL_GLOBAL_VARIABLE );
            }
            else if ( result == EXCLUDE )
            {
                scorepion_service->addMarker( _item, markerList.at( 1 ) );
                if (cubepl_memory!=NULL)
                cubepl_memory->put(variable_adress, index , 0., 0, cube::CUBEPL_GLOBAL_VARIABLE );
            }
        }
    }


    const QList<cubepluginapi::TreeItem*> _all_flattree_items =   scorepion_service->getTreeItems( cubepluginapi::FLATTREE);
    foreach( cubepluginapi::TreeItem * _item, _all_flattree_items )
    {
        cube::Vertex * _vertex = _item->getCubeObject();
        if (_vertex != NULL)
        {
            scorepion_plugin::filters::FilterResult result = filters->apply( _vertex );
            if ( result == EXCLUDE )
                scorepion_service->addMarker( _item, markerList.at( 1 ) );
        }

    }
    if (reduced_trace_size_metric!=NULL )
    {
        scorepion_service->recalculateTreeItem(reduced_trace_size_metric);
        scorepion_service->addMarker( reduced_trace_size_metric, markerList.at( 3 ) );
    }
    if (total_trace_size_metric!=NULL )
    {
        scorepion_service->recalculateTreeItem(total_trace_size_metric);
        scorepion_service->addMarker( total_trace_size_metric, markerList.at( 3 ) );
    }

    if ( (total_trace_size_metric!=NULL) ||
         ( reduced_trace_size_metric!=NULL )
         )
    {
        scorepion_service->updateTreeView(cubepluginapi::METRIC);

    }
    scorepion_service->updateTreeView(cubepluginapi::CALL);
//     scorepion_service->updateTreeView(cubepluginapi::FLATTREE);

    updateMeasurementLabels();


}


void
ScorePionPluginWidget::updateMeasurementLabels()
{
    reduced_trace_size_metric =  scorepion_service->getMetricTreeItem( "total_reduced_trace_size" );
    total_trace_size_metric =  scorepion_service->getMetricTreeItem( "total_full_trace_size" );
	if (reduced_trace_size_metric == NULL || total_trace_size_metric == NULL)
    {
        return;
    }

	reduced_trace_size->setText("mm");
	scorep_total_memory->setText("yy");
	cube::Metric * pure_reduced_trace_size_metric = static_cast<cube::Metric*>(reduced_trace_size_metric->getCubeObject());
	cube::Metric * pure_total_trace_size_metric = static_cast<cube::Metric*>(total_trace_size_metric->getCubeObject());


    cube::IdIndexMap                  metric_id_indices;
    cube::value_container       inclusive_values;
    cube::value_container       exclusive_values;
    cube::list_of_cnodes       cnode_selections;
    const std::vector< cube::Cnode* > &croots = cube->getRootCnodes();
    for( std::vector< cube::Cnode* >::const_iterator citer = croots.begin(); citer!= croots.end(); ++citer )
    {
        cube::list_of_cnodes cnodes;
        cube::cnode_pair     pair;
        pair.first  = *citer;
        pair.second = cube::CUBE_CALCULATE_INCLUSIVE;
        cnode_selections.push_back( pair );
    }
    cube::list_of_sysresources sysres_selection;
    cube->getMetricSubtreeValues(   cnode_selections,
                                    sysres_selection,
                                    *pure_total_trace_size_metric,
                                    static_cast<size_t>(0),
                                    metric_id_indices,
                                    &inclusive_values,
                                    &exclusive_values );

    double _full_trace_size = (inclusive_values[0])?inclusive_values[0]->getDouble():0.;

     cube->getMetricSubtreeValues(  cnode_selections,
                                    sysres_selection,
                                    *pure_reduced_trace_size_metric,
                                    static_cast<size_t>(0),
                                    metric_id_indices,
                                    &inclusive_values,
                                    &exclusive_values );
    double _reduced_trace_size = (inclusive_values[0])?inclusive_values[0]->getDouble():0.;

    QPair<QString, QString> value_unit = scorepion_service->formatNumberAndUnit( _full_trace_size,
		                                                            "bytes",
		                                                            cubegui::FORMAT_TREES,
		                                                            false );
	full_trace_size->setText( value_unit.first + " " + value_unit.second );
	value_unit = scorepion_service->formatNumberAndUnit( _reduced_trace_size,
		                                     "bytes",
		                                     cubegui::FORMAT_TREES,
		                                     false );
	reduced_trace_size->setText( value_unit.first + " " + value_unit.second );

	const std::vector<cube::LocationGroup*>& processes = cube->getLocationGroups();
	double maxMem = 0;


    cube::list_of_metrics       metric_selections;
    cube::metric_pair     mpair;
    mpair.first  = pure_reduced_trace_size_metric;
    mpair.second = cube::CUBE_CALCULATE_INCLUSIVE;
    metric_selections.push_back( mpair );
    cube->getSystemTreeValues(  metric_selections,
                                cnode_selections,
                                inclusive_values,
                                exclusive_values );

	for( std::vector<cube::LocationGroup*>::const_iterator _lg_iter = processes.begin();
		_lg_iter != processes.end(); ++_lg_iter)
	{
		cube::LocationGroup * lg = *_lg_iter;
		double mem = (inclusive_values[ lg->get_sys_id()])? (inclusive_values[ lg->get_sys_id()])->getDouble():0.;
		maxMem = std::max ( maxMem , mem );
	}
	QPair<QString, QString> memory_unit = scorepion_service->formatNumberAndUnit( maxMem,
		                                                            "bytes",
		                                                            cubegui::FORMAT_TREES,
		                                                            false );
	scorep_total_memory->setText( memory_unit.first + " " + memory_unit.second );
    update();
}



void
ScorePionPluginWidget::addedMetric( )
{
    reduced_trace_size_metric =  scorepion_service->getMetricTreeItem( "total_reduced_trace_size" );
    total_trace_size_metric =  scorepion_service->getMetricTreeItem( "total_full_trace_size" );

    cubepl_memory = &(cube->getCubePlMemoryManager());
    variable_adress = cubepl_memory->register_variable( "region_to_filter",  cube::CUBEPL_GLOBAL_VARIABLE);
    region_bytes_adress= cubepl_memory->register_variable( "region_bytes",  cube::CUBEPL_GLOBAL_VARIABLE);
    updateMarkers();
}

void
ScorePionPluginWidget::adjustCountersNum( int c )
{
    if (cubepl_memory != NULL)
    {
        // ----------------- build up list if contributions
#pragma warning (Temporary call, till Plugin turns to Client/Server interface and one doesn't need local Cube instance)

        SCOREP_Score_Profile* profile = new SCOREP_Score_Profile( cube->getCubeObject() );

        SCOREP_Score_Estimator * estimator = new SCOREP_Score_Estimator( profile, c, SCOREP_SCORE_SORTING_TYPE_TOTALTIME );
        const std::vector<cube::Region*> & regs = cube->getRegions();
        uint64_t reg_id = 0;
        for ( std::vector<cube::Region*>::const_iterator iter = regs.begin(); iter != regs.end(); ++iter, ++reg_id)
        {
            cube::Region* r = *iter;
            double index = getRegionId( r );
            int bytes = estimator->bytesPerVisit( reg_id );
            cubepl_memory->put(region_bytes_adress, index , bytes, 0, cube::CUBEPL_GLOBAL_VARIABLE );
        }
        delete profile;
        delete estimator;

        updateMarkers();
    }
    else
    {
        scorepion_service->debug() << "Please add first metrics for OTF2 Trace size estimation" << Qt::endl;
    }
}


double
ScorePionPluginWidget::getRegionId( const cube::Vertex * _c) const
{
    if ( _c == NULL)
        return -1.;

    const cube::Cnode * _cnode = dynamic_cast<const cube::Cnode*>( _c );
    if (_cnode != NULL) // callpath
    {
        return  _cnode->get_callee()->get_id();
    }
    else
    {
        const cube::Region * _region = dynamic_cast<const cube::Region*>( _c );
        if (_region != NULL) // region
        {
            return _region->get_id();
        }
    }
    return -1;
}




void
ScorePionPluginWidget::addIncludeFile( std::vector<cube::Vertex *> & v )
{
    bool create_new = false;
     scorepion_plugin::filters::IncludeFileFilterRulesSet * filter_rules_set = NULL;
    if ( filter_sets.size() == 0)
        create_new = true;
    else
    {
        filter_rules_set = dynamic_cast< scorepion_plugin::filters::IncludeFileFilterRulesSet* >(filter_sets.back()->getFilterRules());
        if (filter_rules_set == NULL)
            create_new= true;
    }


    if (create_new)
    {
        filter_rules_set = new scorepion_plugin::filters::IncludeFileFilterRulesSet();
        ScorePionFilterSetWidget * _set = new ScorePionFilterSetWidget(filter_rules_set);
        connect( _set, SIGNAL( changedFilter ()), this, SLOT(applyFilters ()));
        filter_sets.push_back( _set);
        main_layout->addWidget( _set );
    }


    filter_rules_set = dynamic_cast< scorepion_plugin::filters::IncludeFileFilterRulesSet* >(filter_sets.back()->getFilterRules());

    for (std::vector<cube::Vertex *>::iterator iter = v.begin(); iter != v.end(); ++iter)
    {
        std::string _file = "";
        cube::Vertex * _v = *iter;
        const cube::Cnode * _cnode = dynamic_cast<const cube::Cnode*>( _v );
        if (_cnode != NULL) // callpath
        {
            _file = _cnode->get_callee()->get_mod();
        }
        else
        {
            const cube::Region * _region = dynamic_cast<const cube::Region*>( _v );
            if (_region != NULL) // region
            {
                _file = _region->get_mod();
            }
        }
        filter_rules_set->AddRule( _file);
    }
    ScorePionFilterSetWidget * _w = filter_sets.back();
    _w->rulesChanged();

}


void
ScorePionPluginWidget::addIncludeRegion( std::vector<cube::Vertex *> & v)
{
    bool create_new = false;
     scorepion_plugin::filters::IncludeRegionFilterRulesSet * filter_rules_set = NULL;
    if ( filter_sets.size() == 0)
        create_new = true;
    else
    {
        filter_rules_set = dynamic_cast<scorepion_plugin::filters::IncludeRegionFilterRulesSet*>(filter_sets.back()->getFilterRules());
        if (filter_rules_set == NULL)
            create_new= true;
    }


    if (create_new)
    {
        filter_rules_set = new scorepion_plugin::filters::IncludeRegionFilterRulesSet();
        ScorePionFilterSetWidget * _set = new ScorePionFilterSetWidget(filter_rules_set);
        connect( _set, SIGNAL( changedFilter ()), this, SLOT(applyFilters ()));
        filter_sets.push_back( _set);
        main_layout->addWidget( _set );
    }


    filter_rules_set = dynamic_cast<scorepion_plugin::filters::IncludeRegionFilterRulesSet*>(filter_sets.back()->getFilterRules());

    for (std::vector<cube::Vertex *>::iterator iter = v.begin(); iter != v.end(); ++iter)
    {
        std::string _file = "";
        cube::Vertex * _v = *iter;
        const cube::Cnode * _cnode = dynamic_cast<const cube::Cnode*>( _v );
        if (_cnode != NULL) // callpath
        {
            _file = _cnode->get_callee()->get_name();
        }
        else
        {
            const cube::Region * _region = dynamic_cast<const cube::Region*>( _v );
            if (_region != NULL) // region
            {
                _file = _region->get_name();
            }
        }
        filter_rules_set->AddRule( _file);
    }
    ScorePionFilterSetWidget * _w = filter_sets.back();
    _w->rulesChanged();
}

void
ScorePionPluginWidget::addExcludeFile( std::vector<cube::Vertex *> & v)
{
    bool create_new = false;
     scorepion_plugin::filters::ExcludeFileFilterRulesSet * filter_rules_set = NULL;
    if ( filter_sets.size() == 0)
        create_new = true;
    else
    {
        filter_rules_set = dynamic_cast<scorepion_plugin::filters::ExcludeFileFilterRulesSet*>(filter_sets.back()->getFilterRules());
        if (filter_rules_set == NULL)
            create_new= true;
    }


    if (create_new)
    {
        filter_rules_set = new scorepion_plugin::filters::ExcludeFileFilterRulesSet();
        ScorePionFilterSetWidget * _set = new ScorePionFilterSetWidget(filter_rules_set);
        connect( _set, SIGNAL( changedFilter ()), this, SLOT(applyFilters ()));
        filter_sets.push_back( _set);
        main_layout->addWidget( _set );
    }


    filter_rules_set = dynamic_cast<scorepion_plugin::filters::ExcludeFileFilterRulesSet*>(filter_sets.back()->getFilterRules());

    for (std::vector<cube::Vertex *>::iterator iter = v.begin(); iter != v.end(); ++iter)
    {
        std::string _file = "";
        cube::Vertex * _v = *iter;
        const cube::Cnode * _cnode = dynamic_cast<const cube::Cnode*>( _v );
        if (_cnode != NULL) // callpath
        {
            _file = _cnode->get_callee()->get_mod();
        }
        else
        {
            const cube::Region * _region = dynamic_cast<const cube::Region*>( _v );
            if (_region != NULL) // region
            {
                _file = _region->get_mod();
            }
        }

        filter_rules_set->AddRule( _file);
    }
    ScorePionFilterSetWidget * _w = filter_sets.back();
    _w->rulesChanged();

}

void
ScorePionPluginWidget::addExcludeRegion( std::vector<cube::Vertex *> & v)
{
    bool create_new = false;
     scorepion_plugin::filters::ExcludeRegionFilterRulesSet * filter_rules_set = NULL;
    if ( filter_sets.size() == 0)
        create_new = true;
    else
    {
        filter_rules_set = dynamic_cast<scorepion_plugin::filters::ExcludeRegionFilterRulesSet*>(filter_sets.back()->getFilterRules());
        if (filter_rules_set == NULL)
            create_new= true;
    }


    if (create_new)
    {

        filter_rules_set = new scorepion_plugin::filters::ExcludeRegionFilterRulesSet();
        ScorePionFilterSetWidget * _set = new ScorePionFilterSetWidget(filter_rules_set);
        connect( _set, SIGNAL( changedFilter ()), this, SLOT(applyFilters ()));
        filter_sets.push_back( _set);
        main_layout->addWidget( _set );
    }


    filter_rules_set = dynamic_cast<scorepion_plugin::filters::ExcludeRegionFilterRulesSet*>(filter_sets.back()->getFilterRules());

    for (std::vector<cube::Vertex *>::iterator iter = v.begin(); iter != v.end(); ++iter)
    {
        std::string _file = "";
        cube::Vertex * _v = *iter;
        const cube::Cnode * _cnode = dynamic_cast<const cube::Cnode*>( _v );
        if (_cnode != NULL) // callpath
        {
            _file = _cnode->get_callee()->get_name();
        }
        else
        {
            const cube::Region * _region = dynamic_cast<const cube::Region*>( _v );
            if (_region != NULL) // region
            {
                _file = _region->get_name();
            }
        }
        filter_rules_set->AddRule( _file);
    }
    ScorePionFilterSetWidget * _w = filter_sets.back();
    _w->rulesChanged();
}

 // ----- used by contex menu
void
ScorePionPluginWidget::addIncludeFile( cube::Vertex *  _v)
{
    std::vector<cube::Vertex*> a;
    a.push_back(_v);
    addIncludeFile( a );
}

void
ScorePionPluginWidget::addIncludeRegion( cube::Vertex *  _v)
{
    std::vector<cube::Vertex*> a;
    a.push_back(_v);
    addIncludeRegion( a );
}

void
ScorePionPluginWidget::addExcludeFile( cube::Vertex *  _v)
{
    std::vector<cube::Vertex*> a;
    a.push_back(_v);
    addExcludeFile( a );
}

void
ScorePionPluginWidget::addExcludeRegion( cube::Vertex *  _v)
{
    std::vector<cube::Vertex*> a;
    a.push_back(_v);
    addExcludeRegion( a );
}




    // -- load filter file
void
ScorePionPluginWidget::addFilterRule( scorepion_plugin::filters::FilterRulesSet * filter_rules_set)
{

    ScorePionFilterSetWidget * _set = new ScorePionFilterSetWidget(filter_rules_set);
    connect( _set, SIGNAL( changedFilter ()), this, SLOT(applyFilters ()));
    filter_sets.push_back( _set);
    main_layout->addWidget( _set );
}







void
ScorePionPluginWidget::removeFilterSetWidget( ScorePionFilterSetWidget * widget)
{
    filter_sets.remove( widget );
    main_layout->removeWidget( widget );
    widget->setParent( NULL );
    if (main_layout->count() == 0)
    {
		addDefaultFilterRule();
    }

    applyFilters();

}



    // -- load filter file
void
ScorePionPluginWidget::addDefaultFilterRule( )
{
   std::list<FilterRulesSet *> &f_sets = filter_factory->getDefaultFilterSets();

    if (f_sets.size() != 0)
    {
        for ( std::list<FilterRulesSet *>::iterator iter = f_sets.begin(); iter != f_sets.end(); ++iter)
        {
            FilterRulesSet * _f_set = *iter;
            addFilterRule( _f_set );
        }
    }
 }




void
ScorePionPluginWidget::applyFilters()
{
    filters->clear();
    for (
        std::list<ScorePionFilterSetWidget *>::iterator iter = filter_sets.begin();
        iter != filter_sets.end();
        ++iter
            )
    {
        scorepion_plugin::filters::FilterRulesSet * _filter_set = (*iter)->getFilterRules() ;
        filters->addSet( _filter_set );
    }

    updateMarkers();

}


