/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "config.h"
#include "ScorePionToolBar.h"
#include "PluginServices.h"
#include <QToolButton>
#include <QMessageBox>
#include <QInputDialog>
#include <QMenu>
#include <assert.h>
#include "ScorePionPluginWidget.h"
#include "scorep/SCOREP_Score_Profile.hpp"
#include "scorep/SCOREP_Score_Estimator.hpp"


using namespace scorepion_plugin::gui;

ScorePionToolBar::ScorePionToolBar( cubepluginapi::PluginServices* service ) : QToolBar( "ScorePion Toolbar" )
{
    parent          = 0;
    this->service   = service;

    QAction * open_filter_action = this->addAction(QIcon(":/scorepionicons/open_filter.png"),"Open Filter file");
    connect(open_filter_action, SIGNAL(triggered()),
	    this, SLOT(openFilterFile()));
    QAction * save_filter_action = this->addAction(QIcon(":/scorepionicons/save_filter.png"),"Save Filter file");
    connect(save_filter_action, SIGNAL(triggered()),
	    this, SLOT(saveFilterFile()));

    addSeparator();

    add_incl_filter_action = this->addAction(QIcon(":/scorepionicons/add_to_filter.png"),"Add filter group \"include\"");
    QMenu * incl_filter_menu = new QMenu();
    QAction * incl_file = incl_filter_menu->addAction("Include File");
    connect(incl_file, SIGNAL(triggered()),
	    this, SLOT(addIncludeFile()));

    QAction * incl_region = incl_filter_menu->addAction("Include Region");
    connect(incl_region, SIGNAL(triggered()),
	    this, SLOT(addIncludeRegion()));
    add_incl_filter_action->setMenu( incl_filter_menu);

    add_excl_filter_action = this->addAction(QIcon(":/scorepionicons/remove_from_filter.png"),"Add filter group \"exclude\"");
    QMenu * excl_filter_menu = new QMenu();
    QAction * excl_file = excl_filter_menu->addAction("Exclude File");
    connect(excl_file, SIGNAL(triggered()),
	    this, SLOT(addExcludeFile()));

    QAction * excl_region = excl_filter_menu->addAction("Exclude Region");
    connect(excl_region, SIGNAL(triggered()),
	    this, SLOT(addExcludeRegion()));
    add_excl_filter_action->setMenu( excl_filter_menu);

    addSeparator();
    this->addWidget( new QLabel(tr("Number of counters ")) );
    num_counters_widget = new QSpinBox();
    num_counters_widget->setWhatsThis(tr("Number of additional performance counters to measure."));
    num_counters_widget->setToolTip(tr("Number of additional performance counters to measure."));

    this->addWidget( num_counters_widget );
    connect ( num_counters_widget, SIGNAL(valueChanged( int ) ), this, SLOT(changedCountersNumber( int )) );

    addSeparator();

    QAction * add_trace_estimation_metric_action = this->addAction(QIcon(":/scorepionicons/trace_estimation.png"),"Add OTF2 Trace size estimation metrics");
    connect(add_trace_estimation_metric_action, SIGNAL(triggered()),
	    this, SLOT(addMetric()));



}


void
ScorePionToolBar::setScorePionPluginWidget( ScorePionPluginWidget* newparent )
{
    parent = newparent;
}


// slots connected to the actions in the topology tool bar
void
ScorePionToolBar::showSelf()
{
    if ( parent != NULL )
    {
        this->show();
    }
}

void
ScorePionToolBar::hideSelf()
{
    this->hide();
}


void
ScorePionToolBar::addIncludeFile()
{
    /** Returns selected items in the given tree  @param type METRICTREE, CALLTREE, CALLFLAT, SYSTEMTREE */
    const QList<cubepluginapi::TreeItem*>& selected = service->getSelections( cubepluginapi::CALL );

    std::vector<cube::Vertex * > v;
    foreach ( cubepluginapi::TreeItem* item, selected)
    {
        v.push_back( item->getCubeObject());
    }
    parent->addIncludeFile( v );
}


void
ScorePionToolBar::addIncludeRegion()
{
    /** Returns selected items in the given tree  @param type METRICTREE, CALLTREE, CALLFLAT, SYSTEMTREE */
    const QList<cubepluginapi::TreeItem*>& selected = service->getSelections( cubepluginapi::CALL );

    std::vector<cube::Vertex * > v;
    foreach ( cubepluginapi::TreeItem* item, selected)
    {
        v.push_back( item->getCubeObject());
    }
    parent->addIncludeRegion( v );
}

void
ScorePionToolBar::addExcludeFile()
{
    /** Returns selected items in the given tree  @param type METRICTREE, CALLTREE, CALLFLAT, SYSTEMTREE */
    const QList<cubepluginapi::TreeItem*>& selected = service->getSelections( cubepluginapi::CALL );

    std::vector<cube::Vertex * > v;
    foreach ( cubepluginapi::TreeItem* item, selected)
    {
        v.push_back( item->getCubeObject());
    }
    parent->addExcludeFile( v );
}

void
ScorePionToolBar::addExcludeRegion()
{
    /** Returns selected items in the given tree  @param type METRICTREE, CALLTREE, CALLFLAT, SYSTEMTREE */
    const QList<cubepluginapi::TreeItem*>& selected = service->getSelections(  cubepluginapi::CALL );

    std::vector<cube::Vertex * > v;
    foreach ( cubepluginapi::TreeItem* item, selected)
    {
        v.push_back( item->getCubeObject());
    }
    parent->addExcludeRegion( v );
}


void
ScorePionToolBar::saveFilterFile()
{
    parent->writeFilterFile( );
}


void
ScorePionToolBar::openFilterFile()
{
     parent->loadFilterFile();
}

void
ScorePionToolBar::addMetric()
{
    bool new_metric = false;
    cube::Metric* _total_met  = service->getCube()->getMetric("total_full_trace_size");
    cube::Metric* _reduced_met = service->getCube()->getMetric("total_reduced_trace_size");
    if ( _total_met != NULL )
    {
        service->debug() << "Metric 'Total size of ful  l trace' exists. Ignore. " << Qt::endl;
    }
    else
    {
        new_metric |= true;
         std::string cubepl_init_region_bytes = buildRegionTraceContributionMetric();
        _total_met = service->getCube()->defineMetric(
            "Total size of full trace",
            "total_full_trace_size",
            "INTEGER",
            "bytes",
            "",
            "",
            "",
            NULL,
            cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE,
            "${region_bytes}[${calculation::region::id}]*metric::visits()",
            cubepl_init_region_bytes
        );
      _total_met->setCacheable( false );
      service->addMetric( _total_met, NULL);
    }

    if ( _reduced_met != NULL )
    {
        service->debug() << "Metric 'Total size of reduced trace' exists. Ignore. " << Qt::endl;
    }
    else
    {
        new_metric |= true;
        _reduced_met = service->getCube()->defineMetric(
        "Total size of reduced trace",
        "total_reduced_trace_size",
        "INTEGER",
        "bytes",
        "",
        "",
        "",
        NULL,
        cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE,
        " metric::total_full_trace_size() * ${region_to_filter}[${calculation::region::id} ]"
        );
        _reduced_met->setCacheable( false );
        service->addMetric( _reduced_met, NULL);
    }

    if (new_metric)
        emit addedMetric();
}

std::string
ScorePionToolBar::buildRegionTraceContributionMetric()
{
    // ----------------- build up list if contributions
    cube::CubeProxy*    cube = service->getCube();

#pragma warning ("Temporary call, till Plugin turns to Client/Server interface and one doesn't need local Cube instance")
    SCOREP_Score_Profile* profile = new SCOREP_Score_Profile( cube->getCubeObject() );

    SCOREP_Score_Estimator * estimator = new SCOREP_Score_Estimator( profile, num_counters_widget->value(), SCOREP_SCORE_SORTING_TYPE_TOTALTIME );

    const std::vector<cube::Region*> & regs = cube->getRegions();
    uint64_t reg_id = 0;
    for ( std::vector<cube::Region*>::const_iterator iter = regs.begin(); iter != regs.end(); ++iter, ++reg_id)
    {
        cube::Region* r = *iter;
        bytes_per_visit[ r] = estimator->bytesPerVisit( reg_id );
    }
    delete profile;
    delete estimator;


    std::string cubepl_init_region_bytes= "{\n"
    "global(region_bytes);\n"
    "global(region_to_filter);\n";
    for ( std::map<cube::Region*, uint64_t>::iterator iter = bytes_per_visit.begin(); iter != bytes_per_visit.end(); ++iter)
    {
        std::ostringstream id_out;
        std::ostringstream b_out;
        std::string s_id;
        std::string s_bytes;
        cube::Region* r = iter->first;
        uint64_t id = r->get_id();
        uint64_t b = iter->second;
        id_out << id;
        s_id = id_out.str();
        b_out << b;
        s_bytes = b_out.str();
        cubepl_init_region_bytes += "${region_bytes}[" + s_id + "]=" + s_bytes + "; // " + r->get_name() + "\n";
        cubepl_init_region_bytes += "${region_to_filter}[" + s_id + "]=1;\n";
    }

    cubepl_init_region_bytes +="\n"
    "return 0;\n"
    "}\n";
    return cubepl_init_region_bytes;

}


void
ScorePionToolBar::changedCountersNumber( int i )
{
    emit changedCountersNum( i );
}

