/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <QtGlobal>
#include "FilterRule.h"

#if QT_VERSION <= QT_VERSION_CHECK( 6, 0, 0 )
#include <QRegExp>
#else
#include <QRegularExpression>
#include <QStringView>
#endif

namespace scorepion_plugin
{
    namespace filters
    {

        FilterRule:: FilterRule(const std::string & _regex)
        {
            regex = adjust(_regex);
        }


        std::string
        FilterRule::adjust(const std::string & _v)
        {
            std::string s=_v;
            wildcard_replace( s, ' ', '?');
            wildcard_replace( s, '\t', '?');
            wildcard_replace( s, '\n', '?');
            wildcard_replace( s, ')', '?');
            wildcard_replace( s, '(', '?');
            wildcard_replace( s, '[', '?');
            wildcard_replace( s, ']', '?');
            wildcard_replace( s, '&', '?');
            wildcard_replace( s, '$', '?');
            return s;
        }
        void
        FilterRule::wildcard_replace(std::string & _v, char from, char to) const
        {
            size_t found = std::string::npos;
            while ( (found = _v.find(from)) != std::string::npos)
            {
                _v.replace( found, 1, 1, to);
            }
        }



        bool
        FilterRule:: match_regex(const std::string & _v) const
        {

#if QT_VERSION <= QT_VERSION_CHECK( 6, 0, 0 )
      		QRegExp full_regex (QString::fromStdString( regex), Qt::CaseSensitive, QRegExp::WildcardUnix);
            return full_regex.exactMatch( QString::fromStdString(_v));
#else
            const std::string _r=regex;
            QRegularExpression full_regex (
                    QRegularExpression::wildcardToRegularExpression(QString::fromStdString(_r), QRegularExpression::DefaultWildcardConversion));
            QRegularExpressionMatch match = full_regex.match( QString::fromStdString(_v));
            return match.hasMatch(); // true
#endif

        }



        FilterResult
        FilterRule:: apply(cube::Vertex * _v) const
        {
            (void) _v;
            return NO_MATCH;
        }



        bool
        FilterRule:: match_rule(cube::Vertex * _v) const
        {

            if ( ignore_filtering( _v ))
                return false;
            std::string _str = extract_string( _v );
            return   match_regex( _str );
        }

        bool
        FilterRule:: ignore_filtering(cube::Vertex * _v) const
        {
            std::string _n = "";
            const cube::Cnode * _cnode = dynamic_cast<const cube::Cnode*>( _v );
            if (_cnode != NULL) // callpath
            {
                _n = _cnode->get_callee()->get_paradigm();
            }
            else
            {
                const cube::Region * _region = dynamic_cast<const cube::Region*>( _v );
                if (_region != NULL) // region
                {
                    _n = _region->get_paradigm();
                }
            }
            if (( _n == "compiler" ) || ( _n == "user" ) )
                return false;
            return true;
        }


        void
	    FilterRule::print( std::ostream & out, bool intel)
        {
            if (!intel)
            out << regex;
            else
            {
                std::string star_to_replace = ".*";
                std::string qm_to_replace = ".?";
                std::string to_write = regex;
                size_t found = std::string::npos;
                while ( (found = to_write.find('?', (found == std::string::npos)?0:(found+2))) != std::string::npos)
                {
                    std::string::iterator start = to_write.begin() + found;
                    to_write.replace( start, start + 1, qm_to_replace.begin(), qm_to_replace.end());
                }
                while ( (found = to_write.find('*', (found == std::string::npos)?0:(found+2))) != std::string::npos)
                {
                    std::string::iterator start = to_write.begin() + found;
                    to_write.replace( start, start + 1, star_to_replace.begin(), star_to_replace.end());
                }
                out << to_write;
            }
        };
    };
 };


